#include <windows.h>
#ifdef WINELIB
#include <unix.h>
#include <peexe.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "extrnvar.h"

typedef struct _i386RelocTypes
{
	WORD type;
	PSTR name;
} i386RelocTypes;

// ASCII names for the various relocations used in i386 COFF OBJs
i386RelocTypes i386Relocations[] =
{
{ IMAGE_REL_I386_ABSOLUTE, "ABSOLUTE" },
{ IMAGE_REL_I386_DIR16, "DIR16" },
{ IMAGE_REL_I386_REL16, "REL16" },
{ IMAGE_REL_I386_DIR32, "DIR32" },
{ IMAGE_REL_I386_DIR32NB, "DIR32NB" },
{ IMAGE_REL_I386_SEG12, "SEG12" },
{ IMAGE_REL_I386_SECTION, "SECTION" },
{ IMAGE_REL_I386_SECREL, "SECREL" },
{ IMAGE_REL_I386_REL32, "REL32" }
};
#define I386RELOCTYPECOUNT (sizeof(i386Relocations) / sizeof(i386RelocTypes))

//
// Given an i386 OBJ relocation type, return its ASCII name in a buffer
//
void GetObjRelocationName(WORD type, PSTR buffer, DWORD cBytes)
{
	DWORD i;

	for ( i=0; i < I386RELOCTYPECOUNT; i++ )
		if ( type == i386Relocations[i].type )
		{
			strncpy(buffer, i386Relocations[i].name, cBytes);
			return;
		}

	sprintf( buffer, "???_%X", type);
}

//
// Dump the relocation table for one COFF section
//
/*
Object files contain COFF relocations, which specify how the section data should be
modified when placed in the image file and subsequently loaded into memory.
Image files do not contain COFF relocations, because all symbols referenced have
already been assigned addresses in a flat address space. An image contains
relocation information in the form of base relocations in the .reloc section
(unless the image has the IMAGE_FILE_RELOCS_STRIPPED attribute). See Section 6.5 for
more information.
For each section in an obejct file, there is an array of fixed-length records that
are the sections COFF relocations. The position and length of the array are
specified in the section header. Each element of the array has the following format:
Offset	Size	Field	Description

0		4		Offset	Address of the item to which relocation is applied: this is
						the offset from the beginning of the section, plus the value
					    of the sections RVA/Offset field (see Section 4, 'Section
						Table.'). For example, if the first byte of the section has
						an address of 0x10, the third byte has an address of 0x12.
4		4		Symbol Table Index
						A zero-based index into the symbol table. This symbol gives
					    the address to be used for the relocation. If the specified
						symbol has section storage class, then the symbols address
						is the address with the first section of the same name.
8		2		Type	A value indicating what kind of relocation should be
						performed. Valid relocation types depend on machine type.
						See Section 5.2.1, 'Type Indicators.'

If the symbol referred to (by the Symbol Table Index field) has storage class
IMAGE_SYM_CLASS_SECTION, the symbols address is the beginning of the section. The
section is usually in the same file, except when the object file is part of an
archive (library). In that case, the section may be found in any other object file
in the archive that has the same archive-member name as the current object file.
(The relationship with the archive-member name is used in the linking of import
tables, i.e. the .idata section.)

5.2.1. Type Indicators
The Type field of the relocation record indicates what kind of relocation should be
performed. Different relocation types are defined for each type of machine.

Intel386
The following relocation Type indicators are defined for Intel386 and compatible
processors:

Constant				Value	Description

IMAGE_REL_I386_ABSOLUTE	0		No location is necessary; reference is to an
								absolute value.
IMAGE_REL_I386_DIR16	1		Not supported.
IMAGE_REL_I386_REL16	2		Not supported.
IMAGE_REL_I386_DIR32	3		Direct 32-bit reference to the symbols virtual
								address.
IMAGE_REL_I386_DIR32NB	7		Direct 32-bit reference to the symbols virtual
								address, minus the base. This is a relative virtual
								address.
IMAGE_REL_I386_SEG12	9		Not supported.
IMAGE_REL_I386_SECTION	0xA		Reference to a section address. A common use is to
								facilitate references to debugging information,
								which reside in separate .debug sections.
IMAGE_REL_I386_SECREL	0xB		Reference to an offset from a section address. Has
								same general purpose as IMAGE_REL_I386_SECTION.
IMAGE_REL_I386_REL32	0x14	Program-counter-relative 32-bit reference to the
								symbols virtual address. This is usually a code
								reference.
*/
void DumpObjRelocations(PIMAGE_RELOCATION pRelocs, DWORD count)
{
	DWORD i;
	char szTypeName[32],buffer[128],*szTitle;

	szTitle = "\nAddress  Type    Symbol Index Symbol Name\n-------  ----    ------------ ----- ----\n";
	printf("%s",szTitle);
	for ( i=0; i < count; i++ )
	{
		GetObjRelocationName(pRelocs->Type, szTypeName, sizeof(szTypeName));
		buffer[0] = 0;
        LookupSymbolName(pRelocs->SymbolTableIndex, buffer,sizeof(buffer));

		printf("%-7ld  %-7s   %7ld   %-23s\n",
				pRelocs->VirtualAddress, szTypeName,pRelocs->SymbolTableIndex,buffer);
		if (i > 0 && !(i%25)) printf("%s\n",szTitle);
		pRelocs++;
	}
}

//
// top level routine called from PEDUMP.C to dump the components of a
// COFF OBJ file.
//
void DumpObjFile( PIMAGE_FILE_HEADER pImageFileHeader )
{
	unsigned i;
	PIMAGE_SECTION_HEADER pSections;

	DumpHeader(pImageFileHeader);
	printf("\n");

	pSections = (PIMAGE_SECTION_HEADER)(pImageFileHeader+1);
	PCOFFSymbolTable = MakePtr(PIMAGE_SYMBOL, pImageFileHeader,
								pImageFileHeader->PointerToSymbolTable);

	COFFSymbolCount = pImageFileHeader->NumberOfSymbols;
	DumpSectionTable(pSections, pImageFileHeader->NumberOfSections, FALSE);
	printf("\n");

	if ( fShowRelocations )
	{
		for ( i=0; i < pImageFileHeader->NumberOfSections; i++ )
		{
			if ( pSections[i].PointerToRelocations == 0 )
				continue;

			printf("Section %02d (%.8s) relocations\n", 1+i, pSections[i].Name);
			DumpObjRelocations( MakePtr(PIMAGE_RELOCATION, pImageFileHeader,
									pSections[i].PointerToRelocations),
								pSections[i].NumberOfRelocations );
			printf("\n");
		}
	}


	if ( fShowLineNumbers || fDebug)
	{
		// Walk through the section table...
		for (i=0; i < pImageFileHeader->NumberOfSections; i++)
		{
			// if there's any line numbers for this section, dump'em
			if ( pSections->NumberOfLinenumbers )
			{
				DumpLineNumbers( MakePtr(PIMAGE_LINENUMBER, pImageFileHeader,
										 pSections->PointerToLinenumbers),
								 pSections->NumberOfLinenumbers ,(DWORD)pSections->PointerToLinenumbers);
				printf("\n");
			}
			pSections++;
		}
	}

	if ( fShowSymbolTable || fDebug)
	{
		DumpSymbolTable(PCOFFSymbolTable, COFFSymbolCount);
		printf("\n");
	}

	if ( fShowRawSectionData || fDebug)
	{
		DumpRawSectionData( (PIMAGE_SECTION_HEADER)(pImageFileHeader+1),
							pImageFileHeader,
							pImageFileHeader->NumberOfSections);
	}
}
extern char *ctime(int *);
static char *DumpLibHeader(char *start,int *size,int idx,char *memberName)
{
	int i,siz,j;
	char buf[100],*p;

	printf("\nArchive header:%d\nName:'",idx);
	j = 0;
	for (i=0;i<16;i++) {
		if (start[i] != ' ') memberName[j++] = start[i];
		printf("%c",start[i]);
	}
	printf("'\nDate:'");
	j = 0;
	while (i<16+12) {
//		printf("%c",start[i]);
		if (start[i] != ' ') buf[j++] = start[i];
		i++;
	}
	buf[j] = 0;
	j = atoi(buf);
	p = ctime(&j);
	if (p == NULL) p = "(null)";
	else {
		char *q = p;
		while (*q && *q != '\n') q++;
		if (*q == '\n') *q = 0;
	}
	printf(" %s (%s) ",buf,p);
	printf("'\nUser id:'");
	while (i<16+12+6) {
		printf("%c",start[i]);
		i++;
	}
	printf("'\nGroup id:'");
	while (i<16+12+6+6) {
		printf("%c",start[i]);
		i++;
	}
	printf("'\nFile Mode:'");
	while (i<16+12+6+6+8) {
		printf("%c",start[i]);
		i++;
	}
	printf("'\nFile Size:'");
	j = 0;
	while (i<16+12+6+6+8+10) {
		printf("%c",start[i]);
		buf[j++] = start[i];
		i++;
	}
	buf[j] = 0;
	siz = atoi(buf);
	*size = siz;
	printf("'\n");
	if (i & 1) i++;
	return(start +i);
}
#pragma pack(1)
union aLong {
	char c[4];
	unsigned long l;
};

static long ReadOneBigEndianLong(char *start)
{
	union aLong u;

	u.c[3] = start[0];
	u.c[2] = start[1];
	u.c[1] = start[2];
	u.c[0] = start[3];
	return(u.l);
}

static char *DumpFirstLinkerMember(char *start)
{
	int nrOfSymbols,i,*OffsetsTable;

	printf("\nFirst linker member\n----- ------ ------\n");
	nrOfSymbols = ReadOneBigEndianLong(start);
	start += 4;
	printf("%d (%x) symbols contained in the archive\n",nrOfSymbols,nrOfSymbols);
	OffsetsTable = (int *)malloc(nrOfSymbols * sizeof(int));
	for (i=0; i<nrOfSymbols;i++) {
		OffsetsTable[i] = ReadOneBigEndianLong(start);
		start += 4;
	}
	for (i=0; i<nrOfSymbols;i++) {
		printf("[%4d] %s: Offset %d\n",i,start,OffsetsTable[i]);
		start += strlen(start)+1;
	}
	free(OffsetsTable);
	return(start);
}
static int DumpSecondLinkerMember(char *start)
{
	int nrOfMembers,i,*OffsetsTable,nrOfSymbols;
	unsigned short *SymbolsOffsetsTable;

	printf("\nSecond linker member\n------ ------ ------\n");
	nrOfMembers = *(unsigned long *)start;
	start += 4;
	printf("Number of members: %d\n",nrOfMembers);
	OffsetsTable = (int *)malloc(nrOfMembers * sizeof(int));
	for(i=0; i<nrOfMembers;i++) {
		OffsetsTable[i] = *(unsigned long *)start;
		if (fShowRawSectionData)
			printf("Member %d file offset %d\n",i,OffsetsTable[i]);
		start += 4;
	}
	nrOfSymbols = *(unsigned long *) start;
	start += 4;
	printf("Number of symbols: %d\n",nrOfSymbols);
	SymbolsOffsetsTable = (short *)malloc(nrOfSymbols*sizeof(short));
	for (i=0; i<nrOfSymbols;i++) {
		SymbolsOffsetsTable[i] = *(unsigned short *)start;
		start += 2;
	}
	for (i=0; i<nrOfSymbols;i++) {
		printf("[%4d] %35s %6d %6d\n",i,start,SymbolsOffsetsTable[i],OffsetsTable[SymbolsOffsetsTable[i]-1]);
		start += strlen(start)+1;
	}
	free(OffsetsTable);
	return(nrOfMembers);
}
static char **DumpLongNames(char *start,int siz,int nrOfMembers)
{
	char *stop,**result;
	int i;

	printf("Long names table\n\n");
	stop = start + siz;
	i = 0;
	result = (char **)malloc(nrOfMembers * sizeof(char *));
	while (start < stop) {
		printf("[%4d] '%s'\n",i,start);
		result[i] = start;
		i++;
		start += strlen(start)+1;
	}
	return result;
}
extern int TotalFileSize;

void DumpLibFile(char *start)
{
	char *s,*stop,**LongNamesTable,memberName[18];
	int siz,idx,nrOfMembers;

	stop = start + TotalFileSize;
	s = start;
	start += 8;
	idx = 1;
	LongNamesTable = NULL;
	nrOfMembers = 0;
	do {
		memset(memberName,0,18);
		DumpLibHeader(start,&siz,idx,memberName);
		start += 60;
		if (idx == 1) {
			if (fShowRawSectionData)
				DumpFirstLinkerMember(start);
		}
		else if (idx == 2 &&memberName[0] == '/') {
			nrOfMembers = DumpSecondLinkerMember(start);
		}
		else if (idx == 3 && memberName[0] == '/' && memberName[1] == '/') {
			LongNamesTable = DumpLongNames(start,siz,nrOfMembers);
		}
		else {
			printf("Offset in file %d ",start - s);
			if (LongNamesTable)
			printf("Dump of object file %s. (%s). File offset %d\n",memberName,LongNamesTable[idx-4],start-s);
			else printf("Dump of object file %s\n",memberName);
			DumpObjFile((PIMAGE_FILE_HEADER)start);
		}
		start += siz;
		if (siz & 1) start++;
		idx++;
	} while (start < stop);
}

